"""Parse configurations."""

import sys

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

from dataclasses import dataclass
from pathlib import Path


@dataclass
class Config:
    """Access to configuration values."""

    config_dict: dict[str, dict]
    feature_dict: dict[str, dict]
    feedback_template: dict[str, dict]

    def __init__(self, toml_f_str: str):
        self.config_dict = tomllib.loads(Path(toml_f_str).read_text())
        self.feature_dict = self.config_dict["feature"]
        self.feedback_template = self.config_dict["template"]



from dataclasses import dataclass

from ..config import Config


@dataclass
class EmailConfig:
    """Contain email configurations."""

    account: str
    password: str
    email_from: str
    email_to: str
    smtp_server: str
    subject: str
    email_body: str
    initial_reply: str

    def __init__(self, config: Config):
        config_dict = config.config_dict

        assert "email" in config_dict, "cannot find email configurations in socassess.toml"  # noqa: E501
        assert "account" in config_dict["email"], "cannot find email.account configurations in socassess.toml"  # noqa: E501
        assert "content" in config_dict["email"], "cannot find email.content configurations in socassess.toml"  # noqa: E501

        email_account_dict = config_dict["email"]["account"]
        email_content_dict = config_dict["email"]["content"]

        self.account = email_account_dict["account"]
        self.password = email_account_dict["password"]
        self.email_from = email_account_dict["from"]
        self.email_to = email_account_dict["to"]
        self.smtp_server = email_account_dict["smtp_server"]

        self.subject = email_content_dict["subject"]
        self.email_body = email_content_dict["email_body"]
        self.initial_reply = email_content_dict["initial_reply"]



from .level import FeedbackLevel


def extract(onemap: dict):
    """Extract feedback from a test-feedback map.

    Returns
    -------
    feedback: str

    """
    assert 'feedback' in onemap
    level = onemap['level'] if 'level' in onemap else FeedbackLevel.LOWEST
    feedback = onemap['feedback']
    func_with_params = onemap['function'] if 'function' in onemap else None
    if func_with_params is not None:
        feedback = fill_content(feedback, func_with_params)
    return feedback, level


def fill_content(feedback: str, func_with_params) -> str:
    """Fill {content} in the feedback message."""
    assert func_with_params is not None
    if isinstance(func_with_params, dict):
        func = func_with_params.pop('name')
        params = func_with_params.pop('params')
        feedback = feedback.format(content=func(params=params))
    else:
        # in case there is no params
        func = func_with_params
        feedback = feedback.format(content=func())
    return feedback


def _context(qn, attr):
    """Fetch context given question number (qn) and attribute."""
    if attr is not None and qn in attr:
        attr_context = attr[qn]
    else:
        attr_context = None
    return attr_context


def context(qn, maps):
    """Fetch question, canonical, and student answer contexts.

    All contexts are collected inside the user defined module `maps`.

    """
    ctx = {}
    if "context" in maps.__dict__:
        mapsctx = maps.__dict__["context"]
        for ele in mapsctx:
            _ctx = _context(qn, mapsctx[ele])
            if _ctx is not None:
                ctx |= {
                    ele: _context(qn, mapsctx[ele])
                }
    return ctx



from pathlib import Path
from typing import Final

from ..config import Config
from . import expert as se
from .config import EmailConfig


def attachments(draft_feedback: str,
                artifacts: Path):
    """Collect necessary attachments that should be included in the email.

    Returns
    -------
    attachments: list[dict[str, str]]

    """
    filelist = [
        {
            'data': draft_feedback.encode('utf-8'),
            'maintype': 'text',
            'subtype': 'txt',
            'filename': 'draft_feedback.txt'
        },
    ]
    attachments_list = artifacts / "_attachments.txt"
    if not attachments_list.is_file():
        return filelist

    with attachments_list.open() as flist:
        for line in flist:
            line = line.strip()
            attachment = artifacts / line
            filelist.append(
                {
                    'data': attachment.read_bytes(),
                    'maintype': 'text',
                    'subtype': 'text',
                    'filename': attachment.name + '.txt'
                },
            )
    return filelist


def email_feedback(
        need_expert: dict[str, dict[str, str]],
        current_fb_dict: Final[dict[str, dict[str, str]]],
        ansdir: Path,
        artifacts: Path,
        *,
        config: Config
) -> dict[str, list[str]]:
    """Fallback to human feedback if automated feedback cannot be produced."""
    fb_v = ['\n'.join(x) for x in current_fb_dict.values()]
    current_feedback = '\n---\n'.join(fb_v)

    attachs = attachments(
        current_feedback,
        artifacts,
    )

    email_config = EmailConfig(config)
    se.seek_expert_reply(
        # a list of dict
        attachments=attachs,
        config=email_config
    )

    email_dict = {}
    for qn in need_expert:
        if qn not in current_fb_dict:
            # the actual feedback will be filled later; we just need an empty
            # list here as a placeholder
            email_dict |= {qn: []}
    email_dict |= {'_email': [email_config.initial_reply]}
    return email_dict



import re
from pathlib import Path

import xmltodict

from . import feedback_helper
from .display import formatter, no_auto_feedback
from .exception import SocAssessException
from .level import FeedbackLevel


def generate_feedback(ansdir, artifacts, *, config):
    """Parse test outcomes then produce feedback messages."""
    outcomes = test_outcomes(artifacts)
    fb_dict = automated_feedback(outcomes.keys(), ansdir, artifacts)

    enabled_features: dict[str, bool]
    enabled_features = config.feature_dict

    need_expert = fb_dict.pop('need_expert', None)
    if need_expert is not None:
        # prepare the default feedback
        fb_dict |= no_auto_feedback(need_expert,
                                    template=config.feedback_template)

        # overwrite feedback afterwards
        if "ai" in enabled_features \
           and enabled_features["ai"] is True:
            from .ai import ai_feedback
            fb_dict |= ai_feedback(need_expert, config=config)

        # email has to be after ai so that ai feedback is observable at this
        # point
        if "email" in enabled_features \
           and enabled_features["email"] is True:
            from .email import email_feedback

            # a hidden "_email" key will be added
            fb_dict |= email_feedback(need_expert,
                                      fb_dict,
                                      ansdir,
                                      artifacts,
                                      config=config)

    # reveal raw feedback out of pytests
    if "raw_feedback" in enabled_features \
       and enabled_features["raw_feedback"] is True:
        raw_feedback_list = [
            # None means it is a "passed" test
            x for x in filter(lambda x: x is not None, outcomes.values())
        ]
        if len(raw_feedback_list) > 0:
            fb_dict["_raw_feedback_out_of_testcases"] = raw_feedback_list

    # format feedback dict into text
    # this step has to be the last step
    current_automated_feedback = formatter(fb_dict,
                                           template=config.feedback_template)
    return current_automated_feedback


def test_outcomes(artifacts: Path):
    """Convert xml test outputs into a list of test outcomes.

    test_it::test_a::passed
    test_it::test_b::failed
    ...

    It also saves the message of failed tests and tests raising exceptions,
    which can be displayed if necessary.

    """
    xml = artifacts / 'report.xml'
    outcomes = xmltodict.parse(xml.read_text(encoding='utf-8'))
    testcases = outcomes['testsuites']['testsuite']['testcase']
    if isinstance(testcases, dict):  # if there is only one test outcome
        testcases = [testcases]
    res = {}
    for t in testcases:
        name = t['@classname'] + "::" + t['@name']
        if 'failure' in t:
            outcome = 'failed'
            res[f"{name}::{outcome}"] = t['failure']['@message']
        elif 'error' in t:
            outcome = 'failed'
            res[f"{name}::{outcome}"] = t['error']['@message']
        elif 'skipped' in t:
            continue
        else:
            outcome = 'passed'
            res[f"{name}::{outcome}"] = None
    return res


def automated_feedback(
        outcomes: list,
        ansdir: Path,
        artifacts: Path) -> dict:
    """Map test outcomes into automated feedback messages automatically.

    If automated feedback cannot be found for some questions, then a dict
    containing the context of those questions will be linked with key
    `need_expert`. One can pop that out and handle it afterwards.

    """
    fb_dict = map_feedback(outcomes, ansdir, artifacts)
    return fb_dict


def map_feedback(outcomes: list, ansdir: Path, artifacts: Path):
    """Invoke _map_feedback to generate feedback."""
    try:
        import maps  # dynamically load `maps` under folder `args.feedback`
        fb = _map_feedback(maps, outcomes, ansdir, artifacts)
    except Exception as e:
        raise SocAssessException(e)
    return fb


def _issubset(cur_set, outcomes):
    """Check if current set is a subset of the test outcomes.

    A mapping is typically to build a connection between all pre-coded feedback
    and a large component. A mapping can contain multiple sets, which provides
    pre-coded feedback for small components. Properly joining and/or hiding
    feedback for these small components composes the feedback for the large
    component.

    """
    # step1: handle non regex matches
    if cur_set.issubset(frozenset(outcomes)):
        return True

    # step2: handle regex matches
    for cs in cur_set:
        matched = False
        m = re.match('(.*?)\\[\\*\\]::(passed|failed)', cs, flags=re.I)
        # only regex patterns can be left after step1; otherwise, it means
        # there is no match
        if m is None:
            return False
        pattern = m.group(1)  # prefix pattern
        ot = m.group(2)  # "passed" or "failed"
        for t in outcomes:
            m = re.match(f'{pattern}\\[.*\\]::(passed|failed)', t, flags=re.I)
            # mismatching "passed" or "failed"
            if m is not None:
                matched = True
                if m.group(1) != ot:
                    return False
        if not matched:
            return False
    # we went through all cur_set and every row matches; so the regex matches
    return True


def _map_feedback(maps,
                  outcomes: list, ansdir: Path, artifacts: Path):
    """Provide mappings between test outcomes and feedback messages.

    Additional content can be provided using `{content}`.

    Variables
    ---------
    mappings: dict[set, tuple[function, str, bool]]
        The key is the test outcomes. The value is a dict, containing the
        feedback string. Optionally, it can contain a function to fill
        `{content}` inside the feedback string. Also, a level indicator

    """
    feedback_dict = {}
    for qn in maps.selected:
        cur_mapping = maps.selected[qn]
        qn_dict = {}
        for oneset in cur_mapping:
            if _issubset(oneset, outcomes):
                feedback, fb_level = \
                    feedback_helper.extract(cur_mapping[oneset])
                if fb_level == FeedbackLevel.SINGLE:
                    # stop early in this case
                    return {"_single_feedback_only": [feedback]}
                if fb_level not in qn_dict:
                    qn_dict |= {fb_level: [feedback]}
                else:
                    qn_dict[fb_level].append(feedback)

        if len(qn_dict) == 0:  # if no mapping was found
            if 'need_expert' not in feedback_dict:
                feedback_dict |= {'need_expert': {}}
            qn_context = feedback_helper.context(qn, maps)
            feedback_dict['need_expert'] |= {
                qn: qn_context
            }
        else:
            # only retain the highest feedback level for each question
            qn_highest = qn_dict[sorted(qn_dict.keys(), reverse=True)[0]]
            feedback_dict |= {qn: qn_highest}
    return feedback_dict



from enum import IntEnum


class FeedbackLevel(IntEnum):
    """Provide the grain level of a feedback.

    If higher level feedback exists, lower level feedback will not be shown.

    """

    LOWEST = 10  # default
    LOW = 20
    MEDIUM = 30
    HIGH = 40
    HIGHEST = 50
    # only display this feedback and ignore feedback for all other questions
    SINGLE = 100



import importlib
import shutil
from pathlib import Path

from colorama import Fore


def create_project(name):
    """Set up starter code."""
    assessment_root = Path(name)
    try:
        assessment_root.mkdir(parents=False, exist_ok=False)
    except Exception as e:
        print(e)
        exit(1)

    # set up files under probing_tests/
    probing_tests = assessment_root / 'probing_tests'
    probing_tests.mkdir(parents=False, exist_ok=False)
    (probing_tests / '__init__.py').write_text("")  # empty
    (probing_tests / 'conftest.py').write_text(conftest_content())
    (probing_tests / 'pytest.ini').write_text(pytest_ini_content())
    (probing_tests / 'test_it.py').write_text(test_it_content())

    # set up files under maps/
    maps = assessment_root / 'maps'
    maps.mkdir(parents=False, exist_ok=False)
    (maps / '__init__.py').write_text(maps_init_content())
    (maps / 'general.py').write_text(general_mapping_content())

    # set up files under stu/
    stu = assessment_root / 'stu'
    stu.mkdir(parents=False, exist_ok=False)
    (stu / 'student.txt').write_text(stu_answer_content())

    # set up socassess.toml
    toml_template = importlib.resources.files('socassess') / "socassess.template.toml"  # noqa: E501
    shutil.copy(toml_template, assessment_root / "socassess.toml")

    # done
    starter_messages(name)


def pytest_ini_content():
    """Provide content for pytest.ini file under probing_tests."""
    return """
[pytest]
automark_dependency = True
# always order tests with dependency markers
addopts = --order-dependencies
required_plugins = pytest-timeout pytest-dependency pytest-order
    """.strip()


def conftest_content():
    """Provide content for conftest.py file under probing_tests."""
    return """
from pathlib import Path

import pytest


def pytest_addoption(parser):
    \"\"\"Add necessary pytest configurations.\"\"\"
    parser.addoption(
        "--artifacts", action="store", default="artifacts"
    )
    parser.addoption(
        "--ansdir", action="store", default="stu"
    )


@pytest.fixture(scope="session")
def artifacts(request) -> Path:
    \"\"\"Contains the folder path to store artifacts.\"\"\"
    opt = request.config.getoption("--artifacts")
    return Path(opt)


@pytest.fixture(scope="session")
def stu_answer(request) -> Path:
    \"\"\"Contains the folder path containing student's solution file.\"\"\"
    opt = request.config.getoption("--ansdir")
    return Path(opt) / 'student.txt'


@pytest.fixture(scope="session")
def stu_answer_content(stu_answer) -> Path:
    \"\"\"Read and return the content of the student answer file.\"\"\"
    return stu_answer.read_text()
    """.strip()


def test_it_content():
    """Provide content for test_it.py file under probing_tests/."""
    return """
import shutil

import pytest


def test_exist(stu_answer):
    assert stu_answer.exists()


@pytest.mark.dependency(depends=['test_exist'])
def test_content(stu_answer_content, stu_answer, artifacts):
    \"\"\"Test content only if the solution file exists.\"\"\"
    try:
        assert len(stu_answer_content) > 0
    except AssertionError as e:
        # back up the answer file in case something happens so we can inspect
        # manually later.
        shutil.copy(stu_answer, artifacts)
        raise e
    """.strip()


def maps_init_content():
    """Provide content for __init__.py file under maps/."""
    return """
\"\"\"Contain necessary information for SocAssess to provide feedback..\"\"\"

from . import general

__all__ = [
    # required
    "questions",
    "selected",
]


# ========
# Required
# ========

selected = {
    "general": general.mappings,
}


questions = {
    "general": "this is a general question",
}
    """.strip()


def general_mapping_content():
    """Provide content for general.py file under maps/."""
    return """
mappings = {
    frozenset([
        # Since this is an one-line binary test so we assume the only cause of
        # its failure is the file was not found.
        'test_it::test_exist::failed',
    ]): {
        'feedback': 'Oops! Did you forget to submit the file or did name your file correctly?',  # noqa: E501
    },
    frozenset([
        'test_it::test_content::passed',
    ]): {
        'feedback': 'Nice! Your answer looks good!',
    },
}
    """.strip()


def stu_answer_content():
    """Provide content for student.txt file under stu/."""
    return """
Great! I put something into my answer file.
    """.strip()


def starter_messages(name):
    """Output starter commands."""
    print("Assessment folder is created! Let's give it a try.")
    print(f"    {Fore.GREEN}cd {name}{Fore.RESET}")
    print("To see the automated feedback:")
    print(f"    {Fore.GREEN}socassess feedback --config=socassess.toml --artifacts=artifacts --ansdir=stu --probing=probing_tests --feedback=maps{Fore.RESET}")  # noqa: E501
    print("To inspect pytest outcomes:")
    print(f"    {Fore.GREEN}socassess feedback --artifacts=artifacts --ansdir=stu --probing=probing_tests{Fore.RESET}")  # noqa: E501
    print("Or you can just use `pytest`")
    print(f"    {Fore.GREEN}pytest -vv --artifacts=artifacts --ansdir=stu probing_tests{Fore.RESET}")  # noqa: E501
    print("For more info")
    print(f"    {Fore.GREEN}socassess --help{Fore.RESET}")  # noqa: E501
    print(f"Take a look at the code in {Fore.GREEN}./{name}{Fore.RESET} and have fun!")  # noqa: E501



import smtplib
from email.message import EmailMessage

from .config import EmailConfig


def seek_expert_reply(attachments: list[dict[str, str]] = [],
                      *,
                      config: EmailConfig):
    """Send an email to experts seeking expert feedback.

    Parameters
    ----------
    attachments: list[dict[str, str]]
        A list of attachments. Each is a dict with four keys: `data`,
        `maintype`, `subtype`, `filename`.

    Returns
    -------
    None

    """
    msg = EmailMessage()
    msg['From'] = config.email_from
    msg['To'] = config.email_to
    msg['Subject'] = config.subject
    message = config.email_body.strip()

    msg.set_content(message)

    # handle attachments
    for a in attachments:
        msg.add_attachment(
            a['data'],
            maintype=a['maintype'],
            subtype=a['subtype'],
            filename=a['filename'],
            cte='base64'
        )

    server = smtplib.SMTP(config.smtp_server, 587)
    server.starttls()
    # Login Credentials for sending the mail
    server.login(
       config.account,
       config.password
    )
    # send the message via the server.
    server.sendmail(msg['From'], msg['To'], msg.as_string())
    server.quit()



import contextlib
import io
import os
import shutil
import sys
from pathlib import Path

import pytest

from . import userargs
from .config import Config
from .exception import SocAssessException
from .feedback import generate_feedback
from .parse_args import parse_args


def main() -> int:
    """Provide CLI entry point.

    Raise SocAssessException for development purpose, such exception should
    never be raised during real assessments. For all other Exception, it will
    be silently catched.

    When there is no exception, the exit code is 0; otherwise it will be 1.

    """
    try:
        args = parse_args()

        ansdir = Path(args.ansdir)
        # Always put artifacts onto disk
        artifacts = Path(args.artifacts)

        if args.probing is not None:
            # recreate artifacts only when --probing is used
            shutil.rmtree(artifacts, ignore_errors=True)
            artifacts.mkdir(parents=False, exist_ok=False)
            with Path(args.probing) as pt:
                report_xml = os.path.join(artifacts, 'report.xml')
                # if only "--with-probing", then we output things to console
                # for inspection
                if not args.feedback:
                    pytest.main([
                        f"--junitxml={report_xml}",
                        "-v",
                        "--tb=line",
                        f"--artifacts={artifacts}",
                        f"--ansdir={ansdir}",  # TODO: pytest bug? if
                                               # whitespace appears in previous
                                               # options, remaining options
                                               # cannot be properly parsed
                        pt
                    ])
                else:
                    with io.StringIO() as buf, \
                            contextlib.redirect_stdout(buf), \
                            contextlib.redirect_stderr(buf):
                        pytest.main([
                            f"--junitxml={report_xml}",
                            f"--artifacts={artifacts}",
                            f"--ansdir={ansdir}",
                            pt
                        ])
                        console_output = buf.getvalue()  # noqa: F841

        # allow user to use artifacts
        userargs.artifacts = artifacts
        if args.feedback is not None:
            assert (artifacts / 'report.xml').is_file(), "cannot find report.xml"  # noqa: E501
            assert args.config is not None and Path(args.config).is_file(), \
                "cannot find feedback configuration, e.g., socassess.toml"  # noqa: E501

            # prepare sys.path so we can import `maps` later
            # args.feedback should be named as `maps`
            maps = Path(args.feedback)
            assert maps.name == "maps", "cannot find valid `maps`"
            maps_parent = maps.parent
            # error if there is no parent path
            assert maps_parent != maps, "`maps` should have a parent path"

            sys.path.append(str(maps_parent))

            config = Config(args.config)
            fb = generate_feedback(
                ansdir,
                artifacts,
                config=config
            )
            # always print feedback to stdout
            print(fb)
    except AssertionError as e:
        raise SocAssessException(e) from e
    except SocAssessException as e:
        raise e
    except Exception as e:  # noqa: F841
        print("Oops, something's wrong. Please contact system administrator.")  # noqa: E501
        # for debugging
        # raise e
        return 1



import os
from dataclasses import dataclass

from ..config import Config


@dataclass
class AIConfig:
    """Contain AI configurations."""

    openai_key: str
    model: str
    temperature: float
    max_tokens: int
    top_p: float
    frequency_penaltyre: float
    presence_penalty: float
    system_prompt: str
    feedback_template: str
    textwrap_width: int

    def __init__(self, config: Config):
        config_dict = config.config_dict

        assert "openai" in config_dict, "cannot find ai configurations in socassess.toml"  # noqa: E501

        ai_dict = config_dict["openai"]

        if "openai_key" in ai_dict:
            self.openai_key = ai_dict["openai_key"]
        else:
            self.openai_key = os.environ.get("OPENAI_API_KEY")

        self.model = ai_dict["model"]
        self.temperature = ai_dict["temperature"]
        self.max_tokens = ai_dict["max_tokens"]
        self.top_p = ai_dict["top_p"]
        self.frequency_penalty = ai_dict["frequency_penalty"]
        self.presence_penalty = ai_dict["presence_penalty"]
        self.system_prompt = ai_dict["system_prompt"]
        self.feedback_template = ai_dict["template"]
        self.textwrap_width = ai_dict["textwrap_width"] if "textwrap_width" in ai_dict else None  # noqa: E501



"""Convert feedback dict into text.

Note that the configuration is loaded when the function gets called. This is
because we have to ensure the toml file has been parsed.

"""


def formatter(feedback_dict: dict[str, list[str]], *, template) -> str:
    """Convert feedback dict into text.

    Within a question, the list of its feedback is joined by `one_sep`. It will
    be converted to text using `one_template`. There are two possible keys:
    `question` and `text`.

    Across questions, the list of feedback of all questions is joined by
    `full_sep`. It will be converted to text using `full_template`. There is
    only one possible key: `text`.

    """
    one_sep = template["one_separator"]
    one_template = template["one"]
    full_sep = template["full_separator"]
    full_template = template["full"]

    full_l = []
    for qn in feedback_dict:
        one_text = one_template.format(
            question=qn,
            text=one_sep.join(feedback_dict[qn])
        )
        full_l.append(one_text)
    full_text = full_sep.join(full_l)
    feedback = full_template.format(text=full_text)
    return feedback


def no_auto_feedback(
        need_expert: dict[str, dict[str, str]],
        *,
        template
) -> dict[str, list[str]]:
    """State that no feedback is available.

    If a question does not have automated feedback, then
    `not_available_feedback` will be used. There is only one possible key:
    `question`.

    """
    not_available_feedback = template["not_available"]

    fb_dict = {}
    for qn in need_expert:
        fb_dict |= {qn: [not_available_feedback.format(question=qn)]}
    return fb_dict



import argparse

from .create import create_project


def parse_args():
    """Parse args."""
    parser = argparse.ArgumentParser(
        description="""

A tool to assist assessment creators to create test-based assessments that can
provide high-quality formative feedback

        """.strip())  # noqa: E501
    subparsers = parser.add_subparsers(dest="command")
    # create_parser
    create_parser = subparsers.add_parser('create', help='Create SocAssess starter code in a new folder')  # noqa: E501
    create_parser.add_argument(
        '--name',
        dest='name',
        action='store',
        required=True,
        help='the folder containing the socassess code')

    # feedback_parser
    feedback_parser = subparsers.add_parser('feedback', help='Generate feedback')  # noqa: E501
    feedback_parser.add_argument(
        '--ansdir',
        dest='ansdir',
        action='store',
        required=True,
        default='stu',
        help='the folder containing the answer')
    feedback_parser.add_argument(
        '--artifacts',
        dest='artifacts',
        action='store',
        required=True,
        default='artifacts',
        help='the folder to save generated artifacts, probing test outcomes will be saved to `report.xml`')  # noqa: E501
    feedback_parser.add_argument(
        '--probing',
        dest='probing',
        action='store',
        help='generate artifacts and assess answers with probing tests stored inside the specified folder')  # noqa: E501
    feedback_parser.add_argument(
        '--feedback',
        dest='feedback',
        action='store',
        help='map the test outcomes into feedback messages using mappings inside the specified folder')  # noqa: E501
    feedback_parser.add_argument(
        '--config',
        dest='config',
        action='store',
        # required=True,
        # default='',  # TODO: add a default one
        help='the path to the `socassess.toml` file')  # noqa: E501

    args = parser.parse_args()
    if args.command == 'create':
        create_project(args.name)
        exit(0)
    if args.command == 'feedback' and args.probing is None and args.feedback is None:  # noqa: E501
        feedback_parser.error("Please use at least one of `--probing` or `--feedback`")  # noqa: E501
        exit(1)

    return args



"""Try OpenAI."""

import json
import textwrap

from openai import OpenAI

from ..config import Config
from .config import AIConfig


def ask(client, context, *, config: AIConfig):
    """Ask and get a response."""
    response = client.chat.completions.create(
        model=config.model,
        messages=[
            {
                "role": "system",
                "content": config.system_prompt,
            },
            {
                "role": "user",
                "content": context
            },
        ],
        temperature=config.temperature,
        max_tokens=config.max_tokens,
        top_p=config.top_p,
        frequency_penalty=config.frequency_penalty,
        presence_penalty=config.presence_penalty
    )
    return response


def ai_feedback(
        need_expert: dict[str, dict[str, str]],
        *,
        config: Config
) -> dict[str, list[str]]:
    """Contain logic to seek AI feedback."""
    ai_dict = {}

    ai_config = AIConfig(config)
    client = OpenAI(api_key=ai_config.openai_key)

    for qn, context in need_expert.items():
        response = ask(client, json.dumps(context), config=ai_config)
        ai_fb = response.choices[0].message.content
        if ai_config.textwrap_width is not None:
            ai_fb = textwrap.fill(ai_fb, width=ai_config.textwrap_width)

        ai_dict |= {qn: [ai_config.feedback_template.format(question=qn,
                                                            feedback=ai_fb)]}
    return ai_dict
